% MATLAB script for designing, training, and evaluating a neural network 
% to recognize handwritten digits from images in the MNIST database

% clear all the variables in the Workspace and close all figure windows
clear all
close all

% load data to use for training and testing the network 
load MNIST.mat

% digitInputs is a 784x3000 element matrix, where each column stores the
% contents of one 28x28 pixel image sample - columns of an original 28x28 
% image are stacked one after the next, in one long column of the 
% digitInputs matrix. Each of the 3000 columns stores a different image
% sample. digitTargets is a 10x3000 element matrix in which each column
% represents the correct classification for the corresponding image sample.
% Each column consists of a single 1 in the location whose row index is the 
% true digit, and a 0 at other locations in the column. For example, if
% the 10th image sample (column 10 of the digitInputs matrix) is an image
% of the digit "5", then column 10 of the digitTargets matrix would contain
% the following contents: 0 0 0 0 1 0 0 0 0 0 (with 1 in the 5th row)

displayData(digitInputs', 28);     % display images in a compact grid

% quick trick to create a fully connected neural net (the neural net code 
% prunes links for inputs that have the value 0 in all samples)
prof = digitInputs(:,3000);
prof(prof == 0) = 0.01;
digitInputs(:,3000) = prof;

% change to variable names used in neural net code
x = digitInputs;
t = digitTargets;

% set the number of units in the hidden layer
hiddenLayerSize = 25;

% create a "Pattern Recognition" neural network with the above number
% of hidden units. The second input ('trainscg') specifies that the 
% backpropagation method should be used for training, and the third input
% ('mse') specifies that "Mean Squared Error" should be used to measure
% current performance as the network is trained
net = patternnet(hiddenLayerSize, 'trainscg', 'mse');

% divide the data samples into 3 subsets for training, validation, and 
% testing of the network - the division below enables an equal number of
% data samples for each digit, within each subset

% fraction of total data samples for each subset
trainRatio = 0.7;       
valRatio = 0.15;
testRatio = 0.15;

% vectors to store indices for training, validation, and test samples
trainInd = [];       
valInd = [];
testInd = [];

% calculate number of samples for each subset, for given ratios
nSamples = size(digitInputs,2);
nTargets = size(digitTargets,1);
nTrain = floor(trainRatio * (nSamples/nTargets));
nValid = floor(valRatio * (nSamples/nTargets));
nTest = nSamples - nTrain - nValid;

for i = 1:nTargets                          % for each digit
    inds = find(digitTargets(i,:) == 1);    % get indices for this digit
    trainInd = [trainInd inds(1:nTrain)];   % divide between subsets
    valInd = [valInd inds((nTrain+1):(nTrain+nValid))];
    testInd = [testInd inds((nTrain+nValid+1):end)];
end

% set up function and parameters for network training and testing
net.divideFcn = 'divideind';
net.divideParam.trainInd = trainInd;
net.divideParam.valInd = valInd;
net.divideParam.testInd = testInd;

% parameter influences number of iterations performed during training
net.trainParam.max_fail = 10;

% train the network using the specified inputs and targets
[net,tr] = train(net,x,t);

% test the network on the full set of data samples
y = net(x);
e = gsubtract(t,y);
performance = perform(net,t,y);
tind = vec2ind(t);
yind = vec2ind(y);
netError = sum(tind ~= yind)/numel(tind);
epoch = tr.num_epochs;

% print the percent error and number of iterations in the Command Window
disp(['Percent Error: ' num2str(100*netError) '%'])
disp(['Number of Iterations: ' num2str(epoch)])

% plot the error in network performance as a function of the number of
% iterations during training
figure, plotperform(tr)

% plot a histogram of the number of samples in the training, validation, 
% and test sets whose final error falls within the ranges specified on the
% horizontal axis
figure('Position', [100 100 800 400])
histogram(reshape(e,1,30000), -0.95:0.1:0.95);   % ploterrhist(e)
set(gca, 'XTick', -1:0.1:1);
xlabel('Errors', 'FontWeight', 'bold', 'FontSize', 16)
ylabel('Instances', 'FontWeight', 'bold', 'FontSize', 16)
title('Error Histogram', 'FontWeight', 'bold', 'FontSize', 18)

% plot the confusion matrix that shows the number of samples for which each
% true digit value (shown on the horizontal axis) that were incorrectly 
% classified as the digit shown on the vertical axis
figure, plotconfusion(t,y)

% plot an ROC curve for each digit (1 to 10) in a different color. Each
% ROC curve shows the fraction of image samples that would be correctly 
% classified as this digit, as a function of the number of samples that 
% would be incorrectly classified as this digit, for different choices of 
% the threshold used to decide whether the value computed for the 
% corresponding output unit for this digit should be classified as "true"
figure, plotroc(t,y)



